// pngdib.c
//
// PNGDIB - a mini imaging library for Win32
// By Jason Summers
// This software may be used without restriction.
//
// read_png_to_dib()
//     Read a PNG image file and convert it to a Windows "Device Independent
//     Bitmap" (DIB)
//
// write_dib_to_png()
//     Write a DIB to a PNG image file
//

#include <windows.h>
#include <stdio.h>
#include "png.h"
#include "pngdib.h"

#ifndef uchar
#define uchar unsigned char
#endif


static uchar* uncompress_dib(LPBITMAPINFO lpbmi1, int infosize, void *lpbits1)
{
	LPBITMAPINFOHEADER lpdib2;
	void *lpbits2;
	void *whatever;
	int linesize, bitssize;
	HBITMAP hb;
	HDC hdc;
	HGDIOBJ rvgdi;
	int rvi;
	int width,height;
	LPBITMAPINFOHEADER lpdib1;

	lpdib1=(LPBITMAPINFOHEADER)lpbmi1;
	width=lpdib1->biWidth;
	height=lpdib1->biHeight;

	linesize= (((width * lpdib1->biBitCount)+31)/32)*4;
	bitssize= linesize*height;

	lpdib2=GlobalAlloc(GPTR,infosize);
	if(!lpdib2) return NULL;

	//  create a header for the new uncompressed DIB
	CopyMemory((void*)lpdib2,(void*)lpdib1,infosize);
	lpdib2->biCompression=BI_RGB;
	lpdib2->biSizeImage=0;

	lpbits2= (uchar*)GlobalAlloc(GPTR,bitssize);
	if(!lpbits2) { GlobalFree(lpdib2); return NULL; }


	// Windows bitmap handling functions are not exactly convenient,
	// especially when trying to deal with DIBs. Every function wants
	// to convert them into DDBs. We have to play stupid games and
	// convert back and forth. This probably uses too much memory,
	// and I'm not 100% sure it is exactly correct, but it seems to
	// work for me.

	hb=CreateDIBSection(NULL,(LPBITMAPINFO)lpdib2,DIB_RGB_COLORS,&whatever,NULL,0);

	hdc=CreateCompatibleDC(NULL);
	rvgdi=SelectObject(hdc,hb);
	//SetStretchBltMode(hdc,COLORONCOLOR);
	rvi=StretchDIBits(hdc,
		0,0,width,height,
		0,0,width,height,
		lpbits1, (LPBITMAPINFO)lpdib1,
		DIB_RGB_COLORS,SRCCOPY);
	rvi=GetDIBits(hdc,hb,0,height, lpbits2,
		(LPBITMAPINFO)lpdib2,DIB_RGB_COLORS);

	DeleteDC(hdc);
	DeleteObject(hb);
	GlobalFree(lpdib2);

	return lpbits2;
}


static void p2d_error_fn(png_structp png_ptr, const char *err_msg)
{
	jmp_buf *j;
	j= (jmp_buf*) png_get_error_ptr(png_ptr);
	longjmp(*j, -1);
}


static void p2d_warning_fn(png_structp png_ptr, const char *warn_msg)
{
	return;
}


static void d2p_error_fn(png_structp png_ptr, const char *err_msg)
{
	jmp_buf *j;
	j= (jmp_buf*) png_get_error_ptr(png_ptr);
	longjmp(*j, -1);
}

static void d2p_warning_fn(png_structp png_ptr, const char *warn_msg)
{
	return;
}



int read_png_to_dib(PNGD_P2DINFO *p2dp)
{
	png_structp png_ptr;
	png_infop info_ptr;
	jmp_buf jbuf;
	png_uint_32 width, height;
	int png_bit_depth, color_type, interlace_type;
	png_colorp png_palette;
	png_uint_32 res_x, res_y;
	int res_unit_type;
	FILE *fp;
	int palette_entries;
	uchar **row_pointers;
	uchar *lpdib;
	uchar *dib_palette;
	uchar *dib_bits;
	uchar *tmprow;
	int dib_bpp, dib_bytesperrow;
	int i,j;
	int rv;


	rv=PNGD_E_ERROR;
	png_ptr=NULL;
	fp=NULL;
	row_pointers=NULL;
	lpdib=NULL;

	if(p2dp->structsize != sizeof(PNGD_P2DINFO)) return PNGD_E_VERSION;


	// Set the user-defined pointer to point to our jmp_buf. This will
	// hopefully protect against potentially different sized jmp_buf's in
	// libpng, while still allowing this library to be threadsafe.
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,(void*)(&jbuf),
		p2d_error_fn, p2d_warning_fn);
	if(!png_ptr) return PNGD_E_NOMEM;


	info_ptr = png_create_info_struct(png_ptr);
	if(!info_ptr) {
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		return PNGD_E_NOMEM;
	}


	if(setjmp(jbuf)) {
		// we'll get here if an error occurred in any of the following
		// png_ functions

		rv=PNGD_E_LIBPNG;
		goto abort;
	}

	if((fp = fopen(p2dp->pngfn, "rb")) == NULL) {
		rv=PNGD_E_READ;
		goto abort;
	}

	png_init_io(png_ptr, fp);

	png_read_info(png_ptr, info_ptr);

	png_get_IHDR(png_ptr, info_ptr, &width, &height, &png_bit_depth, &color_type,
		&interlace_type, NULL, NULL);

	switch(color_type) {
	case PNG_COLOR_TYPE_RGB:
	case PNG_COLOR_TYPE_RGB_ALPHA:
		dib_bpp=24;
		palette_entries=0;
		png_set_bgr(png_ptr);
		break;
	case PNG_COLOR_TYPE_PALETTE:
		dib_bpp=png_bit_depth;
		png_get_PLTE(png_ptr,info_ptr,&png_palette,&palette_entries);
		break;
	case PNG_COLOR_TYPE_GRAY:
	case PNG_COLOR_TYPE_GRAY_ALPHA:
		dib_bpp=png_bit_depth;
		if(png_bit_depth>8) dib_bpp=8;
		palette_entries= 1<<dib_bpp;
//		if(png_bit_depth==2) dib_bpp=4;
		// we'll construct a grayscale palette later
		break;
	default:
		rv=PNGD_E_BADPNG;
		goto abort;
	}

	if(dib_bpp==2) dib_bpp=4;

	// We'll overload this var and use 0 to mean no phys data,
	// since if it's not in meters we can't use it anyway
	res_unit_type=0;

	// DIBs *do* support physical resolution
	if(png_get_valid(png_ptr,info_ptr,PNG_INFO_pHYs)) {
		png_get_pHYs(png_ptr,info_ptr,&res_x,&res_y,&res_unit_type);
	}

	png_set_strip_alpha(png_ptr); // DIB's don't support alpha channel
	png_set_strip_16(png_ptr);    // ... or >8 bits per sample

	// DIB scanlines are padded to 4-byte boundaries.
	dib_bytesperrow= (((width * dib_bpp)+31)/32)*4;

	p2dp->dibsize=sizeof(BITMAPINFOHEADER)+4*palette_entries+height*dib_bytesperrow;
	lpdib = (uchar*)GlobalAlloc(GPTR,p2dp->dibsize);
	if(!lpdib) { rv=PNGD_E_NOMEM; goto abort; }
	p2dp->lpdib = (LPBITMAPINFOHEADER)lpdib;

	row_pointers=(uchar**)GlobalAlloc(GMEM_FIXED,height*sizeof(uchar*));
	if(!row_pointers) { rv=PNGD_E_NOMEM; goto abort; }

	// yes, there is some redundancy here
	p2dp->palette_offs=sizeof(BITMAPINFOHEADER);
	p2dp->bits_offs   =sizeof(BITMAPINFOHEADER) + 4*palette_entries;
	dib_palette= &lpdib[p2dp->palette_offs];
	p2dp->palette= (RGBQUAD*)dib_palette;
	dib_bits   = &lpdib[p2dp->bits_offs];
	p2dp->bits = (VOID*)dib_bits;
	p2dp->palette_colors = palette_entries;

	// set up the DIB palette, if needed
	switch(color_type) {
	case PNG_COLOR_TYPE_PALETTE:
		for(i=0;i<palette_entries;i++) {
			p2dp->palette[i].rgbRed   = png_palette[i].red;
			p2dp->palette[i].rgbGreen = png_palette[i].green;
			p2dp->palette[i].rgbBlue  = png_palette[i].blue;
		}
		break;
	case PNG_COLOR_TYPE_GRAY:
	case PNG_COLOR_TYPE_GRAY_ALPHA:
		for(i=0;i<palette_entries;i++) {
			p2dp->palette[i].rgbRed   =
			p2dp->palette[i].rgbGreen =
			p2dp->palette[i].rgbBlue  = (i*255)/(palette_entries-1);
		}
		break;
	}

	for(j=0;j<(int)height;j++) {
		row_pointers[height-1-j]= &dib_bits[j*dib_bytesperrow];
	}

	png_read_image(png_ptr, row_pointers);

	// special handling for this bit depth, since it doesn't exist in DIBs
	// expand 2bpp to 4bpp
	if(png_bit_depth==2) {
		tmprow = GlobalAlloc(GPTR, (width+3)/4 );
		if(!tmprow) { rv=PNGD_E_NOMEM; goto abort; }

		for(j=0;j<(int)height;j++) {
			CopyMemory(tmprow, row_pointers[j], (width+3)/4 );
			ZeroMemory(row_pointers[j], (width+1)/2 );

			for(i=0;i<(int)width;i++) {
				row_pointers[j][i/2] |=
					( ((tmprow[i/4] >> (2*(3-i%4)) ) & 0x03)<< (4*(1-i%2)) );
			}
		}
		GlobalFree(tmprow);
	}

	GlobalFree((VOID*)row_pointers);
	row_pointers=NULL;

	png_read_end(png_ptr, info_ptr);

	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	png_ptr=NULL;

	fclose(fp);
	fp=NULL;

	// fill in the DIB header fields
	p2dp->lpdib->biSize=          sizeof(BITMAPINFOHEADER);
	p2dp->lpdib->biWidth=         width;
	p2dp->lpdib->biHeight=        height;
	p2dp->lpdib->biPlanes=        1;
	p2dp->lpdib->biBitCount=      dib_bpp;
	p2dp->lpdib->biCompression=   BI_RGB;  // uncompressed
	// biSizeImage can also be 0 in uncompressed bitmaps
	p2dp->lpdib->biSizeImage=     height*dib_bytesperrow;
	if(res_unit_type==1) {
		p2dp->lpdib->biXPelsPerMeter= res_x;
		p2dp->lpdib->biYPelsPerMeter= res_y;
	}
	p2dp->lpdib->biClrUsed=       palette_entries;
	p2dp->lpdib->biClrImportant=  0;

	return PNGD_E_SUCCESS;

abort:

	if(png_ptr) png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	if(fp) fclose(fp);
	if(row_pointers) GlobalFree(row_pointers);
	if(lpdib) GlobalFree(lpdib);

	return rv;
}

//--------------------------------------------------------------------

int write_dib_to_png(PNGD_D2PINFO *d2pp)
{
	png_structp png_ptr;
	png_infop info_ptr;
	jmp_buf jbuf;
	png_text text_ptr[1];
	uchar *bits;
	uchar *newimage;
	RGBQUAD* dib_palette;
	int headersize, dib_bpp;
	png_uint_32 res_x, res_y;
	png_color_8 pngc8;
	int height,width;
	int palette_entries;
	int topdown;
	int dib_bytesperrow;
	int compression;
	FILE *fp;
	png_color png_palette[256];
	uchar **row_pointers;
	int i,x,y,size;
	DWORD *bitfields;
	unsigned int v;
	int bf_format;   // bitfields format identifier
	int iscompressed;
	int palsize;  // bytes in palette or bitfields

	int palentsize;  // bytes in a palette entry: 3 or 4;
	BITMAPCOREHEADER *lpolddib;
	int rv;  // return code


	rv=PNGD_E_ERROR;  // this should always get changed before returning
	png_ptr=NULL;
	fp=NULL;
	row_pointers=NULL;
	newimage=NULL;

	if(d2pp->structsize != sizeof(PNGD_D2PINFO)) return PNGD_E_VERSION;

	headersize= d2pp->lpdib->biSize;

	if(headersize<40 && headersize!=12) return PNGD_E_BADBMP;

	if(headersize==12) {
		// This is to support an old kind of DIBs (OS/2) that aren't really
		// used anymore.
		palentsize= 3;
		lpolddib= (BITMAPCOREHEADER*) d2pp->lpdib;
		width= lpolddib->bcWidth;
		height= lpolddib->bcHeight;
		dib_bpp= lpolddib->bcBitCount;
		compression = BI_RGB;
		res_x = res_y = 0;

		// This will get adjusted later if there is a palette.
		// Not sure it's right, though. Do old DIBs always have a
		// full-sized palette?
		palette_entries=0;
	}
	else {
		palentsize=4;
		width= d2pp->lpdib->biWidth;
		height= d2pp->lpdib->biHeight;
		dib_bpp= d2pp->lpdib->biBitCount;
		compression= d2pp->lpdib->biCompression;
		res_x = d2pp->lpdib->biXPelsPerMeter;
		res_y = d2pp->lpdib->biYPelsPerMeter;
		palette_entries = d2pp->lpdib->biClrUsed;
	}

	// supposedly, if the height is negative, the top scanline is stored first
	topdown=0;
	if(height<0) {
		height= -height;
		topdown=1;
	}

	// sanity check
	if(height<1 || height>1000000 || width<1 || width>1000000)
		return PNGD_E_BADBMP;

	// only certain combinations of compression and bpp are allowed
	switch(compression) {
	case BI_RGB:
		if(dib_bpp!=1 && dib_bpp!=4 && dib_bpp!=8 && dib_bpp!=16
			&& dib_bpp!=24 && dib_bpp!=32) return PNGD_E_UNSUPP;
		break;
	case BI_RLE4:
		if(dib_bpp!=4) return PNGD_E_UNSUPP;
		break;
	case BI_RLE8:
		if(dib_bpp!=8) return PNGD_E_UNSUPP;
		break;
	case BI_BITFIELDS:
		if(dib_bpp!=16 && dib_bpp!=32) return PNGD_E_UNSUPP;
		break;
	default:
		return PNGD_E_UNSUPP;
	}

	iscompressed= (compression==BI_RLE4 || compression==BI_RLE8);

	// uncompressed dibs are padded to 4-byte bondaries
	dib_bytesperrow= (((width * dib_bpp)+31)/32)*4;

	if(dib_bpp>8) {
		palette_entries=0;
	}
	else {
		// 0 means it has a full palette
		if(palette_entries==0) palette_entries = 1<<dib_bpp;
	}

	if(compression==BI_BITFIELDS) palsize=12;
	else palsize=palentsize*palette_entries;

	// bounds check
	size= headersize+palsize;
	if(d2pp->dibsize) {
		if(size>d2pp->dibsize) return PNGD_E_BADBMP;
	}

	if(d2pp->lpbits) {
		if(d2pp->bitssize && !iscompressed) {   // bounds check
			size=dib_bytesperrow*height;
			if(size>d2pp->bitssize) return PNGD_E_BADBMP;
		}

		bits=(uchar*)d2pp->lpbits;
	}
	else {
		// If not provided by user, assume the bits immediately
		// follow the palette.

		if(d2pp->dibsize && !iscompressed) {  // bounds check
			size= headersize+palsize+dib_bytesperrow*height;
			if(size>d2pp->dibsize) return PNGD_E_BADBMP;
		}

		bits= &((uchar*)(d2pp->lpdib))[headersize+palsize];
	}

	dib_palette = (RGBQUAD*) ( &((uchar*)(d2pp->lpdib))[headersize] );

	bitfields=(DWORD*)dib_palette;

	bf_format=0;
	if(compression==BI_BITFIELDS) {
		if(dib_bpp==16) {
			if     (bitfields[0]==0x00007c00 && bitfields[1]==0x000003e0 &&
			        bitfields[2]==0x0000001f) bf_format=11;  // 555
			else if(bitfields[0]==0x0000f800 && bitfields[1]==0x000007e0 &&
			        bitfields[2]==0x0000001f) bf_format=12;  // 565
			else { return PNGD_E_UNSUPP; }
		}
		if(dib_bpp==32) {
			if     (bitfields[0]==0x00ff0000 && bitfields[1]==0x0000ff00 &&
			        bitfields[2]==0x000000ff) bf_format=21;
			else { return PNGD_E_UNSUPP; }
		}
	}

	if(bf_format==0 && dib_bpp==16) bf_format=10;
	if(bf_format==0 && dib_bpp==32) bf_format=20;


	// Done analyzing the DIB, now time to convert it to PNG

	// jbuf: see comments in read_png_to_dib()
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (void*)(&jbuf),
	    d2p_error_fn, d2p_warning_fn);
	if (!png_ptr) return PNGD_E_NOMEM;


	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
		png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
		return PNGD_E_NOMEM;
	}

	if(setjmp(jbuf)) {
		// we'll get here if an error occurred in any of the following
		// png_ functions
		rv=PNGD_E_LIBPNG;
		goto abort;
	}

	fp=fopen(d2pp->pngfn,"wb");
	if(!fp) {
		rv=PNGD_E_WRITE;
		goto abort;
	}

	png_init_io(png_ptr, fp);

	png_set_IHDR(png_ptr, info_ptr, width, height, (dib_bpp>8)?8:dib_bpp,
		(dib_bpp>8)?PNG_COLOR_TYPE_RGB:PNG_COLOR_TYPE_PALETTE,
		(d2pp->flags&PNGD_INTERLACED)?PNG_INTERLACE_ADAM7:PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	// For 16-bit DIBs, we get to write an sBIT chunk.
	if(bf_format==10 || bf_format==11) {
		pngc8.red= 5;  pngc8.green= 5;  pngc8.blue= 5;
		png_set_sBIT(png_ptr, info_ptr, &pngc8);
	}
	if(bf_format==12) {
		pngc8.red= 5;  pngc8.green= 6;  pngc8.blue= 5;
		png_set_sBIT(png_ptr, info_ptr, &pngc8);
	}

	if(res_x>0 && res_y>0)
		png_set_pHYs(png_ptr, info_ptr, res_x, res_y, 1);


	if(palette_entries>0) {
		for(i=0;i<palette_entries;i++) {
			if(palentsize==3) {
				png_palette[i].red   = ((RGBTRIPLE*)dib_palette)[i].rgbtRed;
				png_palette[i].green = ((RGBTRIPLE*)dib_palette)[i].rgbtGreen;
				png_palette[i].blue  = ((RGBTRIPLE*)dib_palette)[i].rgbtBlue;
			}
			else {
				png_palette[i].red   = dib_palette[i].rgbRed;
				png_palette[i].green = dib_palette[i].rgbGreen;
				png_palette[i].blue  = dib_palette[i].rgbBlue;
			}
		}
		png_set_PLTE(png_ptr, info_ptr, png_palette, palette_entries);
	}

	if(dib_bpp>8)
		png_set_bgr(png_ptr);

	png_write_info(png_ptr, info_ptr);

	row_pointers=(uchar**)GlobalAlloc(GMEM_FIXED,height*sizeof(uchar*));
	if(!row_pointers) { rv=PNGD_E_NOMEM; goto abort; }


	if(dib_bpp==16 || dib_bpp==32) {

		// Special handling for these bit depths.
		// This uses a lot of memory, and could be improved by processing
		// one line at a time (but that makes it tricky to write interlaced
		// images).

		newimage=(uchar*)GlobalAlloc(GMEM_FIXED,height*width*3);
		if(!newimage) { rv=PNGD_E_NOMEM; goto abort; }

		for(y=0;y<height;y++) {
			for(x=0;x<width;x++) {
				switch(bf_format) {
				case 10:  case 11:  // 16-bit, format 555 (xRRRRRGG GGGBBBBB)
					v= bits[y*dib_bytesperrow+x*2+0] | (bits[y*dib_bytesperrow+x*2+1]<<8);
					newimage[(y*width+x)*3+0]= (v & 0x0000001f)<<3; // blue
					newimage[(y*width+x)*3+1]= (v & 0x000003e0)>>2; // green
					newimage[(y*width+x)*3+2]= (v & 0x00007c00)>>7; // red
					break;
				case 12:            // 16-bit, format 565 (RRRRRGGG GGGBBBBB)
					v= bits[y*dib_bytesperrow+x*2+0] | (bits[y*dib_bytesperrow+x*2+1]<<8);
					newimage[(y*width+x)*3+0]= (v & 0x0000001f)<<3; // blue
					newimage[(y*width+x)*3+1]= (v & 0x000007e0)>>3; // green
					newimage[(y*width+x)*3+2]= (v & 0x0000f800)>>8; // red
					break;
				case 20:  case 21:  // 32-bit, every 4th byte wasted (b g r x)
					newimage[(y*width+x)*3+0]= bits[y*dib_bytesperrow+x*4+0]; // blue
					newimage[(y*width+x)*3+1]= bits[y*dib_bytesperrow+x*4+1]; // green
					newimage[(y*width+x)*3+2]= bits[y*dib_bytesperrow+x*4+2]; // red
					break;
				}
			}
		}

		for(i=0;i<height;i++) {
			if(topdown)
				row_pointers[i]= &newimage[i*width*3];
			else
				row_pointers[height-1-i]= &newimage[i*width*3];
		}
		png_write_image(png_ptr, row_pointers);

		GlobalFree(newimage);
		newimage=NULL;
	}
	else if(iscompressed) {
		newimage= uncompress_dib((LPBITMAPINFO)d2pp->lpdib, headersize+palsize, bits);
		if(!newimage) { rv=PNGD_E_NOMEM; goto abort; }
		for(i=0;i<height;i++) {
			if(topdown)
				row_pointers[i]= &newimage[i*dib_bytesperrow];
			else
				row_pointers[height-1-i]= &newimage[i*dib_bytesperrow];
		}
		png_write_image(png_ptr, row_pointers);

		GlobalFree(newimage);
		newimage=NULL;
	}
	else {
		for(i=0;i<height;i++) {
			if(topdown)
				row_pointers[i]= &bits[i*dib_bytesperrow];
			else
				row_pointers[height-1-i]= &bits[i*dib_bytesperrow];
		}
		png_write_image(png_ptr, row_pointers);
	}

	GlobalFree((VOID*)row_pointers);
	row_pointers=NULL;

	if(d2pp->software) {
		text_ptr[0].key = "Software";
		text_ptr[0].text = d2pp->software;
		text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
		png_set_text(png_ptr, info_ptr, text_ptr, 1);
	}

	png_write_end(png_ptr, info_ptr);
	rv=PNGD_E_SUCCESS;

abort:
	if(png_ptr) png_destroy_write_struct(&png_ptr, &info_ptr);
	if(fp) fclose(fp);
	if(row_pointers) GlobalFree(row_pointers);
	if(newimage) GlobalFree(newimage);

	return rv;
}
